

#include <linux/profiler.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <string.h>

/*
 * Dumbomarbitrary limits
 */

#define LINE_LIMIT 100
#define SYSMAP_LIMIT 10000

struct trace_entry * trace_table;

struct sysmap_entry {
	unsigned long int pc;
	char * name;
};

struct sysmap_entry sysmap_table [SYSMAP_LIMIT];

int sysmap_entries = 0;
int MHZ=100;

void read_sysmap (void)
{
	unsigned long pc;
	char type;
	int i, len;

	FILE * file;
	char str [LINE_LIMIT+1];
	
	file=fopen("System.map","r");

	if (!file) {
		fprintf(stderr,"System.map missing.\n");
		exit(-1);
	}

	sysmap_table[0].pc = 0;
	sysmap_table[0].name = "<address smaller than first Sysmap entry>\n";

	sysmap_entries=1;

	while (fscanf(file, "%lx %1c", &pc, &type) == 2) {
		i=sysmap_entries++;
		if (!fgets(str, LINE_LIMIT, file)) {
			perror("ouch, System.map format error.\n");
			exit(-1);
		}
		sysmap_table[i].pc = pc;
		sysmap_table[i].name = malloc(LINE_LIMIT);
		if (!sysmap_table[i].name) {
			perror("ouch, outta mem.\n");
			exit(-1);
		}
		/*
		 * Dirty trick to strip off end of line:
		 */
		len = strlen(str);
		str[len-1]=0;
		strcpy (sysmap_table[i].name, str);
	}

	printf("read %d lines from System.map.\n", sysmap_entries-1);

	sysmap_table[sysmap_entries].pc = ~0;
	sysmap_table[sysmap_entries].name = "<address larger than last Sysmap entry>\n";
	sysmap_entries++;

/*
 *	for (i=0; i<sysmap_entries-1; i++)
 *		printf("read: %lx %s", sysmap_table[i].pc, sysmap_table[i].name);
 */

	fclose(file);
}

int match_pc (unsigned long pc)
{
	int first=0,last=sysmap_entries-1, middle;

	/*
	 * Not exactly perfect binary searching ... but we never have exact matching
	 * anyways.
	 */
	while (last-first>1) {
		middle = first+(last-first)/2;
		if (sysmap_table[middle].pc <= pc)
			first = middle;
		else
			last = middle;
	}

	return first;
}

void read_proc_info (void)
{
	int bytes,size;
	int i;
	unsigned int prev_stamp,timestamp;
	unsigned int min_latency;

	int file;
	
	file=open("/proc/trace",O_RDONLY);

	if (!file) {
		perror("/proc/trace missing.\n");
		exit(-1);
	}

	read (file, &size, sizeof(int));
	lseek(file,4,SEEK_SET);

	trace_table=(struct trace_entry *)malloc(size*sizeof(struct trace_entry));

	bytes=read(file,trace_table,size*sizeof(struct trace_entry));

	if (size*sizeof(struct trace_entry)!=bytes) {
		printf("something went wrong, bytes read: %d, tried: %d.\n",bytes,size*sizeof(struct trace_entry));
		exit(-1);
	}

	prev_stamp=0;

	/*
	 * Pass 1: look for ~0 which signals calibration latencies.
	 */

	min_latency = ~0;
	for (i=0; i<size; i++) {

		if (trace_table[i].pc == ~0) {
			static int flag=0;
			unsigned int delta;
			delta = trace_table[i].timestamp-trace_table[i-1].timestamp;

			if (delta < min_latency)
				min_latency=delta;

			if (!flag) {
				int idx_prev = match_pc(trace_table[i-1].pc);

				flag=1;

				if (strcmp(sysmap_table[idx_prev].name," read_trace"))
					fprintf(stderr,"hmm, possibly wrong System.map?.\n");
			}
		}
	}

	if (min_latency == ~0) {
		fprintf(stderr,"huh, no calibration entries found?.\n");
		fprintf(stderr,"using 0.39 usecs.\n");
		min_latency = 0.39*MHZ;
	} else
		printf("calibration done, estimated measurement latency: %3.2f microseconds.\n\n", min_latency/(double)MHZ);

	min_latency -= 10;

	prev_stamp = trace_table[0].timestamp;

	for (i=1; i<size-1; i++) {
		unsigned int delta, idx;

		timestamp = trace_table[i+1].timestamp;

		delta = timestamp-prev_stamp;

		idx = match_pc(trace_table[i].pc);

/*		printf("%08ld\n",timestamp);*/
		printf("%08lx %s +<%lx/%lx> (%3.02f)\n",
			trace_table[i].pc,
			sysmap_table[idx].name,
			trace_table[i].pc-sysmap_table[idx].pc,
			sysmap_table[idx+1].pc - sysmap_table[idx].pc,
			(delta-min_latency)/(double)MHZ);

		prev_stamp=timestamp;

	}

	free(trace_table);
	close(file);

	printf("\n");
}


int main(int argc, char * * argv)
{
	if (	   (argc != 3)
	 	|| (strcmp(argv[1], "--cpu")))
	{
		fprintf(stderr,"usage: ktrace --cpu MHZ\n");
		exit (-1);
	}

	if ((!sscanf(argv[2], "%d", &MHZ)) || (MHZ<0) || (MHZ>1000)) {
		fprintf(stderr,"problem parsing MHZ.\n");
		exit (-1);
	}

	printf("MHZ: %d.\n",MHZ);

	read_sysmap();
	read_proc_info();
	return 0;
}


